import re

class gfaNode:
    def __init__(self, name="", sequence="", length=0, contig="", offset=0, sr=0):
        self.name = name
        self.sequence = sequence
        self.len = length
        self.contig = contig
        self.offset = offset
        self.sr = sr

def read_gfa(ref_graph):
    """Parse a GFA file to extract node and edge information."""
    gfa_node, gfa_edge = {}, {}
    with open(ref_graph, "r") as fp:
        for line in fp:
            tokens = line.strip().split('\t')
            if tokens[0] == "S":  # Sequence line
                name = tokens[1]
                sequence = tokens[2]
                try:
                    length = int(tokens[3][5:])  # e.g., LN:i:1234 -> 1234
                    contig = tokens[4][5:]  # e.g., RC:Z:chr1 -> chr1
                    offset = int(tokens[5][5:])  # e.g., OF:i:456 -> 456
                    sr = int(tokens[6][5:])  # e.g., SR:i:1
                except (IndexError, ValueError):
                    raise ValueError(f"Invalid S-line format in GFA: {line}")

                gfa_node[name] = gfaNode(name, sequence, length, contig, offset, sr)

            # elif tokens[0] == "L":  # Link/edge line
            #     gfa_edge.setdefault(tokens[1], []).append(line)

    return gfa_node

def analyze_cigar_indel(tuples, min_length, is_gaf=False):
    """
    Parses CIGAR tuples and returns indels with length >= min_length.
    If is_gaf=True, expects GAF-style tuples (length, op).
    Otherwise, expects BAM-style tuples (op, length).
    """
    pos_ref = 0
    pos_read = 0
    indels = []

    # BAM CIGAR code to string mapping
    bam_op_map = {
        0: 'M',  # match
        1: 'I',  # insertion
        2: 'D',  # deletion
        4: 'S',  # soft clip
        7: '=',  # match
        8: 'X',  # mismatch
    }

    for t in tuples:
        if is_gaf:
            length, op = t
        else:
            op_code, length = t
            op = bam_op_map.get(op_code, None)
            if op is None:
                continue  # skip unknown op codes

        if op == 'M' or op == '=' or op == 'X':
            pos_ref += length
            pos_read += length
        elif op == 'I':
            if length >= min_length:
                indels.append((pos_ref, pos_read, length, "INS"))
            pos_read += length
        elif op == 'D':
            if length >= min_length:
                indels.append((pos_ref, pos_read, length, "DEL"))
            pos_ref += length
        elif op == 'S':
            pos_read += length

    return indels

def merge_cigar(sigs, read_type, min_indel_length=300):
    """
       Merge nearby SV signatures of the same type if they are within max_merge distance
       and each is longer than min_indel_length.
       """
    i = 0
    max_merge = 1000 if read_type == 'hifi' else 500
    while i < len(sigs) - 1:
        diff = abs(sigs[i + 1].start - sigs[i].start)
        if (sigs[i].type == sigs[i + 1].type and diff <= max_merge and sigs[i].svlen >= min_indel_length and sigs[i + 1].svlen >= min_indel_length):
            sigs[i].end += sigs[i + 1].svlen
            sigs[i].svlen += sigs[i + 1].svlen
            sigs.pop(i + 1)
        else:
            i += 1

    return sigs

def sorted_nicely(vcf_entries):
    """ Sort the given vcf entries (in the form ((contig, start, end), vcf_string, sv_type)) in the way that humans expect.
        e.g. chr10 comes after chr2
        Algorithm adapted from https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/"""
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    tuple_key = lambda entry: (alphanum_key(str(entry[0][0])), entry[0][1], entry[0][2])
    return sorted(vcf_entries, key=tuple_key)

# class ToolMissingError(Exception): pass
#
# def check_prereqisites():
#     tools = {
#         'bgzip': ['bgzip', '--version'],
#         'tabix': ['tabix', '--version'],
#         'minigraph': ['minigraph', '--version'],
#         'bcftools': ['bcftools', '--version'],
#         'truvari': ['truvari', '--version']
#     }
#
#     for tool, cmd in tools.items():
#         try:
#             subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
#         except (subprocess.CalledProcessError, FileNotFoundError) as e:
#             raise ToolMissingError(
#                 f"The pipeline cannot be started because '{tool}' is missing or failed to run.") from e


class Signature:
    """Signature class for basic signatures of structural variants. An signature is always detected from a single read.
    """
    def __init__(self, contig, start, length, signature, read_name):
        self.contig = contig
        self.start = start
        self.end = start + length
        self.signature = signature
        self.read_name = read_name
        self.svlen = self.end - self.start

    def get_source(self):
        return (self.contig, self.start, self.end)

    def get_key(self):
        return self.get_source()

    def downstream_distance_to(self, signature2):
        """Return distance >= 0 between this signature's end and the start of signature2."""
        this_contig, this_start, this_end = self.get_source()
        other_contig, other_start, other_end = signature2.get_source()
        if self.type == signature2.type and this_contig == other_contig:
            return max(0, other_start - this_end)
        else:
            return float("inf")

class SignatureDeletion(Signature):
    """SV Signature: a region (contig:start-end) has been deleted and is not present in sample"""

    def __init__(self, contig, start, length, signature, read_name, read_seq=None, pos_read=None, node_ls=None):
        self.contig = contig
        # assert end >= start
        # 0-based start of the deletion (first deleted base)
        self.start = start
        # 0-based end of the deletion (one past the last deleted base)
        self.end = start+length
        self.signature = signature
        self.read_name = read_name
        self.type = "DEL"
        self.svlen = length
        self.pos_read = pos_read
        self.read_seq = read_seq
        self.node_ls = node_ls

class SignatureInsertion(Signature):
    """SV Signature: a region of length end-start has been inserted at contig:start"""

    def __init__(self, contig, start, length, signature, read_name, read_seq=None, pos_read=None, alt_seq=None, qry_start=None, qry_end=None, node_ls=None):
        self.contig = contig
        # assert end >= start
        # 0-based start of the insertion (base after the insertion)
        self.start = start
        # 0-based start of the insertion (base after the insertion) + length of the insertion
        self.end = start + length
        self.signature = signature
        self.read_name = read_name
        self.type = "INS"
        self.svlen = length
        self.pos_read = pos_read
        self.read_seq = read_seq
        self.alt_seq = alt_seq
        self.qry_start = qry_start
        self.qry_end = qry_end
        self.node_ls = node_ls

    def get_key(self):
        contig, start, end = self.get_source()
        return (self.type, contig, start)

    def downstream_distance_to(self, signature2):
        """Return distance >= 0 between this signature's end and the start of signature2."""
        this_contig, this_start, this_end = self.get_source()
        other_contig, other_start, other_end = signature2.get_source()
        if self.type == signature2.type and this_contig == other_contig:
            return max(0, other_start - this_start)
        else:
            return float("inf")

class SignatureInversion(Signature):
    """SV Signature: a region (contig:start-end) has been inverted in the sample"""

    def __init__(self, contig, start, end, signature, read, direction):
        self.contig = contig
        assert end >= start
        # 0-based start of the inversion (first inverted base)
        self.start = start
        # 0-based end of the inversion (one past the last inverted base)
        self.end = end
        self.signature = signature
        self.read = read
        self.type = "INV"
        self.direction = direction
        self.svlen = self.end - self.start

class SignatureDuplicationTandem(Signature):
    """SV Signature: a region (contig:start-end) has been tandemly duplicated"""

    def __init__(self, contig, start, end, signature, read):
        self.contig = contig
        assert end >= start
        # 0-based start of the region (first copied base)
        self.start = start
        # 0-based end of the region (one past the last copied base)
        self.end = end
        self.signature = signature
        self.read = read
        self.type = "DUP"
        self.svlen = self.end - self.start

class SignatureTranslocation(Signature):
    """SV Signature: two positions (contig1:pos1 and contig2:pos2) are connected in the sample"""

    def __init__(self, contig1, pos1, direction1, contig2, pos2, direction2, signature, read):
        if contig1 < contig2 or (contig1 == contig2 and pos1 < pos2):
            self.contig1 = contig1
            # 0-based source of the translocation (first base before the translocation)
            self.pos1 = pos1
            self.source_direction = direction1
            self.contig2 = contig2
            # 0-based destination of the translocation (first base after the translocation)
            self.pos2 = pos2
            self.dest_direction = direction2
        else:
            print('translocation error')
        self.signature = signature
        self.read = read
        self.type = "BND"
        self.svlen = None

    def get_source(self):
        return (self.contig1, self.pos1, self.pos1 + 1)

    def get_destination(self):
        return (self.contig2, self.pos2, self.pos2 + 1)

    def get_key(self):
        return (self.type, self.contig1, self.pos1)





import re
import sys
import os
import pickle
import logging
import time
from multiprocessing import Pool
import pysam
from time import strftime, localtime
import subprocess

from svpg.input_parsing import parse_arguments
from svpg.SVCollect import read_bam
from svpg.SVCluster import form_bins, cluster_data
from svpg.SVPan import read_gaf, read_gaf_pan
from svpg.util import read_gfa
from svpg.output_vcf import consolidate_clusters_unilocal, write_final_vcf
from svpg.SVGenotype import genotype
from svpg.graph_augment import augment_pipe


options = parse_arguments()
if options.sub == 'call':
    bam = pysam.AlignmentFile(options.bam, threads=options.num_threads)
ref_genome = pysam.FastaFile(options.ref)

def multi_process(total_len, step, args=None):
    num_threads = int(options.num_threads)
    chunk_size = total_len // num_threads

    analysis_pools = Pool(processes=int(num_threads))
    async_results = []
    for i in range(num_threads):
        start = i * chunk_size
        end = start + chunk_size if i < num_threads - 1 else total_len
        if step == 'read_bam':
            async_results.append(analysis_pools.starmap_async(read_bam, [(args, start, end, options)]))
        elif step == 'read_gaf':
            async_results.append(analysis_pools.starmap_async(read_gaf, [(args[0][start: end], args[1], options)]))
        elif step == 'read_gaf_pan':
            async_results.append(analysis_pools.starmap_async(read_gaf_pan, [(args[0][start: end], args[1], options)]))
        elif step == 'cluster':
            async_results.append(analysis_pools.starmap_async(cluster_data, [(args[0][start:end], args[1])]))
        else:
            async_results.append(analysis_pools.starmap_async(genotype, [(args[0][start:end], args[1], options)]))

    analysis_pools.close()
    analysis_pools.join()
    results = []
    for async_result in async_results:
        result = async_result.get()
        results.extend(result)
    return [item for sublist in results for item in sublist]

def read_in_chunks(file_object, chunk_size=102400):
    while True:
        lines = []
        for _ in range(chunk_size):
            line = file_object.readline().decode('utf-8').strip()
            if not line:
                break
            lines.append(line)
        if not lines:
            break
        yield lines

def main():
    # Set up logging
    logFormatter = logging.Formatter("%(asctime)s [%(levelname)-7.7s]  %(message)s")
    rootLogger = logging.getLogger()
    rootLogger.setLevel(logging.INFO)

    # Ensure the base directory exists
    os.makedirs(options.working_dir, exist_ok=True)

    # Create log file
    fileHandler = logging.FileHandler(
        "{0}/SVPG_{1}.log".format(options.working_dir, strftime("%y%m%d_%H%M%S", localtime())), mode="w")
    fileHandler.setFormatter(logFormatter)
    rootLogger.addHandler(fileHandler)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)
    logging.info("******************************** Start SVPG ******************************")

    logging.info("CMD: python3 {0}".format(" ".join(sys.argv)))
    logging.info("WORKING DIR: {0}".format(os.path.abspath(options.working_dir)))
    for arg in vars(options):
        logging.info("PARAMETER: {0}, VALUE: {1}".format(arg, getattr(options, arg)))

    if options.sub == 'call' or options.sub == 'graph-call':
        if not options.min_support:
            try:
                if options.read == 'ont' or options.read == 'clr':
                    options.min_support = options.depth // 10 + 2
                else:
                    options.min_support = options.depth // 10
            except AttributeError:
                print(
                    "Please specify the min_support or sequencing depth.")
                return

        gfa_node = read_gfa(options.gfa)

        if options.contigs is None:
            options.contigs = [ctg for ctg in ref_genome.references if re.match(r'^(chr)?[0-9XYM]+$', ctg)]

    pan_signatures = []
    if options.sub == 'call':
        logging.info("MODE: call")
        logging.info("INPUT: {0}".format(os.path.abspath(options.bam)))
        logging.info("***************** Collect SV signatures *****************")

        try:
            bam.check_index()
        except ValueError:
            logging.warning(
                "Input BAM file is missing a valid index. Please generate with 'samtools faidx'.")
        except AttributeError:
            logging.warning(
                "pysam's check_index raised an Attribute error. Something is wrong with the input BAM file.")
            return

        bam_signatures = []
        ref_list = bam.get_index_statistics()
        for ref in ref_list:
            if ref.mapped == 0:
                continue
            if ref[0] in options.contigs:
                ref_len = bam.get_reference_length(ref[0])
                logging.info("Processing ref {0}...".format(ref[0]))
                bam_signatures.extend(multi_process(ref_len, 'read_bam', ref[0]))
                logging.info("Processed ref {0}...".format(ref[0]))
        #
        logging.info("****************************** Graph Mapping ******************************")

        # with open(options.working_dir + '/sv_signatures.pkl', 'wb') as temp:
        #     pickle.dump(bam_signatures, temp)
        # with open(options.working_dir +'/sv_signatures.pkl', 'rb') as f:
        #     bam_signatures = pickle.load(f)
        if not os.path.exists(options.working_dir + '/signatures.fa'):
            fasta_file = open(options.working_dir + '/signatures.fa', 'a')
            signature_bin, bin_depth = form_bins(bam_signatures, 1000)
            signature_bin_remap = [cluster for cluster in signature_bin if len(cluster) >= options.min_support]
            for cluster in signature_bin_remap:
                for sig in cluster:
                    pos_ref = str(sig.contig) + ':' + str(sig.start) + ':' + str(sig.end)
                    # adjac_distance = max(min(5000, sig.svlen*3), 2000)
                    adjac_distance = 2000
                    if sig.signature == 'suppl':
                        read_seq = sig.read_seq
                    elif sig.signature == 'merged_indel':
                        read_seq = sig.read_seq
                    else:
                        if sig.pos_read < adjac_distance:
                            read_seq = sig.read_seq[0:sig.pos_read + sig.svlen + adjac_distance]
                        else:
                            read_seq = sig.read_seq[sig.pos_read - adjac_distance:sig.pos_read + sig.svlen + adjac_distance] if sig.pos_read else sig.read_seq

                    ref_suppl1, ref_suppl2 = '', ''
                    if sig.pos_read < adjac_distance:
                        try:
                            ref_suppl1 = ref_genome.fetch(sig.contig, sig.start - adjac_distance, sig.start - sig.pos_read)
                        except ValueError:
                            ref_suppl1 = ''
                        sig.signature = 'short'
                    if sig.type == 'DEL' and sig.pos_read + adjac_distance > len(sig.read_seq):    # 窗口右端超出read末端
                        try:
                            ref_suppl2 = ref_genome.fetch(sig.contig, sig.end + (len(sig.read_seq) - sig.pos_read),
                                                          sig.end + adjac_distance)
                        except ValueError:
                            ref_suppl2 = ref_genome.fetch(sig.contig, sig.end + (len(sig.read_seq) - sig.pos_read),
                                                          len(ref_genome.fetch(sig.contig)))
                        sig.signature = 'short'
                    elif sig.type == 'INS' and sig.pos_read + sig.svlen + adjac_distance > len(sig.read_seq):
                        try:
                            ref_suppl2 = ref_genome.fetch(sig.contig, sig.start + len(sig.read_seq) - sig.pos_read - sig.svlen,
                                                      sig.start + adjac_distance)
                        except ValueError:
                            ref_suppl2 = ref_genome.fetch(sig.contig, sig.start + len(sig.read_seq) - sig.pos_read - sig.svlen,
                                                          len(ref_genome.fetch(sig.contig)))
                        sig.signature = 'short'
                    read_seq = ref_suppl1 + read_seq + ref_suppl2
                    fasta_file.write(f'>{sig.read_name}@{sig.type}@{pos_ref}@{sig.signature}\n{read_seq}\n')
            fasta_file.close()

            if options.read == 'hifi':
                os.system(
                        'minigraph -t 128 -cx asm --vc --secondary yes  {0} {1}/signatures.fa > {1}/signatures.gaf'.format(options.gfa, options.working_dir))
            else:
                os.system(
                        'minigraph -t 128 -cx lr --vc --secondary yes {0} {1}/signatures.fa > {1}/signatures.gaf'.format(options.gfa, options.working_dir))

        logging.info("*************** Collect signatures from pangenome-reference ***************")

        with open(options.working_dir + '/signatures.gaf', 'rb') as f:
            for chunk_index, lines in enumerate(read_in_chunks(f, chunk_size=200000000)):
                logging.info(f"Processing chunks {chunk_index + 1}")
                pan_signatures.extend(multi_process(len(lines), 'read_gaf', (lines, gfa_node)))
                logging.info(f"Processed chunks {chunk_index + 1}")

    elif options.sub == 'graph-call':
        logging.info("MODE: graph-call")
        logging.info("INPUT: {0}".format(os.path.abspath(options.gaf)))
        logging.info("*************** Collect SV signatures from pangenome ***************")

        with open(options.gaf, 'rb') as f:
            for chunk_index, lines in enumerate(read_in_chunks(f, chunk_size=200000000)):
                logging.info(f"Processing chunk {chunk_index+1}")
                pan_signatures.extend(multi_process(len(lines), 'read_gaf_pan', (lines, gfa_node)))
                # pan_signatures.extend(read_gaf_pan(lines, gfa_node,options))
                logging.info(f"Processed chunks {chunk_index+1}")

    elif options.sub == 'augment':
        logging.info("MODE: augment")
        logging.info("*************** Collect SVs from pangenome ***************")
        start_time = time.time()

        base_dir = options.working_dir
        if not options.skip_call:
            filelist_path = os.path.join(base_dir, "filelist.tsv")
            if os.path.exists(filelist_path):
                os.remove(filelist_path)

            sample_paths_to_process = []
            if options.sample_list:
                try:
                    with open(options.sample_list, 'r') as f:
                        sample_paths_to_process = [line.strip() for line in f if line.strip()]
                except FileNotFoundError:
                    logging.error(f"Sample list file not found: {options.sample_list}")
                    raise RuntimeError(f"Sample list file not found: {options.sample_list}")
            else:
                # Fallback to directory scanning if sample_list is not provided
                for entry in os.scandir(base_dir):
                    if entry.is_dir() and entry.name.startswith("HG"):
                        # Assuming FASTA file is directly inside the sample directory and named as {prefix}.fasta
                        fasta_path_in_dir = os.path.join(entry.path, f"{entry.name}.fasta")
                        if os.path.exists(fasta_path_in_dir):
                            sample_paths_to_process.append(fasta_path_in_dir)
                        else:
                            logging.warning(
                                f"Expected FASTA file {fasta_path_in_dir} not found. Skipping {entry.name}.")
                            raise RuntimeError(f"FASTA file not found for sample: {entry.name}")

            with open(filelist_path, "a") as filelist:
                for fasta_file_path in sample_paths_to_process:
                    try:
                        # Get the directory of the fasta file and its prefix
                        sample_dir = os.path.dirname(fasta_file_path)
                        prefix = os.path.basename(sample_dir) if sample_dir else \
                        os.path.splitext(os.path.basename(fasta_file_path))[0]

                        original_cwd = os.getcwd()
                        os.chdir(sample_dir)

                        fasta_file_name = os.path.basename(fasta_file_path)

                        logging.info(f"Start call SVs from sample {prefix}")
                        file_size = os.path.getsize(fasta_file_name)
                        coverage = file_size // (1024 * 1024 * 1024) // 3.1
                        hifi_support_map = [
                            (0, 5, 1), (5, 15, 2), (15, 25, 3), (25, 50, 4), (50, float("inf"), 5)
                        ]
                        ont_support_map = [
                            (0, 5, 2), (5, 15, 3), (15, 25, 4), (25, 50, 5), (50, float("inf"), 10)
                        ]
                        support_map = hifi_support_map if options.read == 'hifi' else ont_support_map
                        support = next(val for low, high, val in support_map if low <= coverage <= high)

                        gaf_file = f"{prefix}.gaf"
                        if not os.path.exists(gaf_file):
                            if options.read == 'hifi':
                                cmd_align = f"minigraph -t128 -cxasm --vc --secondary yes {options.gfa} {fasta_file_name} > {gaf_file}"
                            else:
                                cmd_align = f"minigraph -t128 -cxlr --vc --secondary yes {options.gfa} {fasta_file_name} > {gaf_file}"
                            subprocess.run(cmd_align, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE,
                                           text=True)

                        var_file = options.vcf_out
                        cmd_call = [
                            "python", __file__, "graph-call",
                            "--read", options.read,
                            "-s", str(support),
                            "--working_dir", './',  # This should refer to the current sample directory
                            "--ref", options.ref,
                            "--gfa", options.gfa,
                            "--gaf", gaf_file,
                            "-o", var_file,
                            "--raw_fasta", fasta_file_name,
                            "--min_sv_size", str(options.min_sv_size),
                            "--max_sv_size", str(options.max_sv_size)
                        ]
                        try:
                            subprocess.run(cmd_call, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE, text=True)
                            pysam.tabix_compress(var_file, f"{var_file}.gz", force=True)
                            pysam.tabix_index(f"{var_file}.gz", preset="vcf", force=True)
                        except subprocess.CalledProcessError:
                            logging.error(
                                f"'{prefix}' encountered an error while running the SVs call.\
                                Please check the logs in the sample directory: {sample_dir}"
                            )
                            raise RuntimeError(f"Error occurred for sample: {prefix}") from e

                        # write the VCF path to filelist.tsv
                        vcf_path = os.path.join(sample_dir, f"{var_file}.gz")
                        if os.path.exists(vcf_path):
                            filelist.write(f"{vcf_path}\n")

                    except Exception as e:
                        logging.error(f"Failed to process sample from path {fasta_file_path}: {e}")
                    finally:
                        # Always change back to the original working directory
                        os.chdir(original_cwd)

        call_time = time.time()
        logging.info(f"SVs call time: {call_time - start_time:.2f} seconds")

        logging.info("*************** Augment pangenome graph ***************")
        augment_pipe(base_dir, options.ref, options.gfa, options.out)
        end_time = time.time()
        logging.info(f"Graph augment time: {end_time - call_time:.2f} seconds")
        logging.info(f"Total time: {end_time - start_time:.2f} seconds")

        return

    # with open(options.working_dir + '/pan_signatures.pkl', 'wb') as temp:
    #     pickle.dump(pan_signatures, temp)
    # with open(options.working_dir+'/pan_signatures.pkl', 'rb') as f:
    #     pan_signatures = pickle.load(f)
    deletion_signatures = [ev for ev in pan_signatures if ev.type == "DEL"]
    insertion_signatures = [ev for ev in pan_signatures if ev.type == "INS"]
    logging.info("Found {0} signatures for deleted regions.".format(len(deletion_signatures)))
    logging.info("Found {0} signatures for inserted regions.".format(len(insertion_signatures)))

    logging.info("**************************** Cluster signatures ***************************")

    filted_signature = []
    bin_depth_list = []
    for element_signature in [deletion_signatures, insertion_signatures]:
        if not element_signature:
            continue
        signature_bin, bin_depth = form_bins(element_signature, 1000)
        bin_depth_list.append(bin_depth)
        signature_clusters = []
        signature_clusters.extend(multi_process(len(signature_bin), 'cluster', (signature_bin, bin_depth)))
        entropies, node_ls_collection = [], []
        for cluster in signature_clusters:
            entropie_flag = 0
            for current, next_item in zip(cluster[:-1], cluster[1:]):
                if current.node_ls and next_item.node_ls:
                    if set(current.node_ls) != set(next_item.node_ls):
                        entropie_flag = 1
                        break
                else:
                    if current.svlen != next_item.svlen or current.start != next_item.start:
                        entropie_flag = 1
                        break
            entropies.append(entropie_flag)
        for ne, cluster in zip(entropies, signature_clusters):
            if len(cluster) >= options.min_support:
                filted_signature.append(cluster)
            else:
                if options.min_support > 2:
                    if options.min_support - 1 == len(cluster) and ne == 0:
                        filted_signature.append(cluster)

        logging.info(f"Generated {len(signature_clusters)} signature clusters")

    chrom_results = {}
    for sig in filted_signature:
        chrom_results.setdefault(sig[0].contig, []).append(sig)

    logging.info("********************************** SVCALL *********************************")

    sv_candidate = []
    fasta_file = pysam.FastaFile(options.raw_fasta) if getattr(options, "raw_fasta", None) else None

    for contig in chrom_results:
        if contig in options.contigs:
            ref_chrom_seq = ref_genome.fetch(contig)
            sv_candidate.extend(sorted(
                consolidate_clusters_unilocal(chrom_results[contig], ref_chrom_seq, options, fasta_file),
                key=lambda cluster: (cluster.contig, cluster.start)))

    deletion_candidates = [i for i in sv_candidate if i.type == 'DEL']
    insertion_candidates = [i for i in sv_candidate if i.type == 'INS']
    logging.info("Final deletion candidates: {0}".format(len(deletion_candidates)))
    logging.info("Final insertion candidates: {0}".format(len(insertion_candidates)))

    if options.sub == 'call' and not options.skip_genotype:
        logging.info("********************************* GENOTYPE ********************************")
        logging.info("Genotyping deletions..")
        deletion_candidates = multi_process(len(deletion_candidates), 'genotype', (deletion_candidates, "DEL"))
        logging.info("Genotyping insertions..")
        insertion_candidates = multi_process(len(insertion_candidates), 'genotype', (insertion_candidates, "INS"))

    types_to_output = [entry.strip() for entry in options.types.split(",")]
    write_final_vcf(deletion_candidates,
                    insertion_candidates,
                    ref_genome.references,
                    ref_genome.lengths,
                    types_to_output,
                    options.working_dir,
                    options.out)

if __name__ == "__main__":
    try:
        sys.exit(main())
    except Exception as e:
        logging.error(e, exc_info=True)


import logging
import pysam

from svpg.SVSignature import SignatureDeletion, SignatureInsertion
from svpg.util import analyze_cigar_indel, merge_cigar

def decompose_cigars(alignment, bam, query_name, min_length):
    """Parse BAM record to extract SVs from CIGAR."""
    sv_signatures = []
    ref_chr = bam.getrname(alignment.reference_id)
    ref_start = alignment.reference_start
    indels = analyze_cigar_indel(alignment.cigartuples, min_length)
    read_seq = alignment.query_sequence
    for pos_ref, pos_read, length, typ in indels:
        start = ref_start + pos_ref
        if typ == "DEL":
            sv_signatures.append(SignatureDeletion(ref_chr, start, length, "cigar", query_name, read_seq=read_seq, pos_read=pos_read))
        elif typ == "INS":
            sv_signatures.append(SignatureInsertion(ref_chr, start, length, "cigar", query_name, read_seq=read_seq, pos_read=pos_read))

    return sv_signatures

def analyze_split_indel(alignment_current, alignment_next, ultra_ins_flag=False):
    """Parse BAM record to extract SVs from inter-alignment."""
    distance_on_read = alignment_next['q_start'] - alignment_current['q_end']
    if not alignment_current['is_reverse']:
        distance_on_reference = alignment_next['ref_start'] - alignment_current['ref_end']
        if alignment_next['is_reverse']:  # INV:+-
            if alignment_current['ref_end'] > alignment_next['ref_end']:
                distance_on_reference = alignment_next['ref_end'] - alignment_current['ref_start']
            else:
                distance_on_reference = alignment_current['ref_end'] - alignment_next['ref_start']
    else:
        distance_on_reference = alignment_current['ref_start'] - alignment_next['ref_end']
        if not alignment_next['is_reverse']:  # INV:-+
            if alignment_current['ref_end'] > alignment_next['ref_end']:
                distance_on_reference = alignment_next['ref_end'] - alignment_current['ref_start']
            else:
                distance_on_reference = alignment_current['ref_end'] - alignment_next['ref_start']

    split_signature = []
    deviation = distance_on_read - distance_on_reference
    if alignment_current['ref_chr'] == alignment_next['ref_chr']:
        if alignment_current['is_reverse'] == alignment_next['is_reverse']:
            if distance_on_reference >= -50:
                # INS
                if deviation >= 50:
                    if not alignment_current['is_reverse']:
                        if not ultra_ins_flag:
                            start = alignment_current['ref_end']
                        else:
                            start = min(alignment_current['ref_end'], alignment_next['ref_start'])
                        pos_read = alignment_current['q_end']
                    else:
                        if not ultra_ins_flag:
                            start = alignment_current['ref_start']
                        else:
                            start = min(alignment_current['ref_start'], alignment_next['ref_end'])
                        pos_read = alignment_next['infer_read_length'] - alignment_next['q_start']
                    split_signature.append(SignatureInsertion(
                        alignment_current['ref_chr'], start, deviation, "suppl", alignment_current['read_name'], read_seq=alignment_current['atgc_seq'], pos_read=pos_read, alt_seq='<INS>'))
                # DEL
                elif deviation <= -50:
                    if not alignment_current['is_reverse']:
                        start = alignment_current['ref_end']
                        pos_read = alignment_current['q_end']
                    else:
                        start = alignment_next['ref_end']
                        pos_read = alignment_next['infer_read_length'] - alignment_next['q_start']
                    split_signature.append(SignatureDeletion(alignment_current['ref_chr'], start, -deviation, "suppl", alignment_current['read_name'], read_seq=alignment_current['atgc_seq'], pos_read=pos_read))
                else:
                    return split_signature

    return split_signature

def decompose_split(primary, supplementaries, bam):
    """Parse BAM record to extract SVs from split_reads."""
    read_name = primary.query_name
    alignments = [primary] + supplementaries
    alignment_list = []
    sig_list = []
    for alignment in alignments:
        if alignment.is_reverse:
            q_start = alignment.infer_read_length() - alignment.query_alignment_end
            q_end = alignment.infer_read_length() - alignment.query_alignment_start
        else:
            q_start = alignment.query_alignment_start
            q_end = alignment.query_alignment_end

        alignment_dict = {
            'read_name': read_name,
            'q_start': q_start,
            'q_end': q_end,
            'ref_chr': bam.getrname(alignment.reference_id),
            'ref_start': alignment.reference_start,
            'ref_end': alignment.reference_end,
            'is_reverse': alignment.is_reverse,
            'mapping_quality': alignment.mapping_quality,
            'infer_read_length': alignment.infer_read_length(),
            'atgc_seq': primary.query_sequence,
        }
        alignment_list.append(alignment_dict)

    sorted_alignment_list = sorted(alignment_list, key=lambda aln: (aln['q_start'], aln['q_end']))
    for index in range(len(sorted_alignment_list) - 1):
        sig_list.extend(analyze_split_indel(sorted_alignment_list[index], sorted_alignment_list[index + 1]))
    if len(alignment_list) >= 3 and sorted_alignment_list[0]['ref_chr'] != sorted_alignment_list[1]['ref_chr']:
        sig_list.extend(
            analyze_split_indel(sorted_alignment_list[0], sorted_alignment_list[-1], ultra_ins_flag=True))

    return sig_list

def retrieve_other_alignments(main_alignment, bam):
    """Reconstruct other alignments of the same read for a given alignment from the SA tag"""
    if main_alignment.get_cigar_stats()[0][5] > 0:
        return []
    try:
        sa_tag = main_alignment.get_tag("SA").split(";")
    except KeyError:
        return []
    other_alignments = []
    # For each other alignment encoded in the SA tag
    for element in sa_tag:
        # Read information from the tag
        fields = element.split(",")
        if len(fields) != 6:
            continue
        rname = fields[0]
        pos = int(fields[1])
        strand = fields[2]
        # CIGAR string encoded in SA tag is shortened
        cigar = fields[3]
        mapq = int(fields[4])
        nm = int(fields[5])

        # Generate an aligned segment from the information
        a = pysam.AlignedSegment()
        a.query_name = main_alignment.query_name
        a.query_sequence = main_alignment.query_sequence
        if strand == "+":
            a.flag = 2048
        else:
            a.flag = 2064
        a.reference_id = bam.get_tid(rname)
        a.reference_start = pos - 1
        try:
            a.mapping_quality = mapq
        except OverflowError:
            a.mapping_quality = 0
        a.cigarstring = cigar
        a.next_reference_id = -1
        a.next_reference_start = -1
        a.template_length = 0
        a.query_qualities = main_alignment.query_qualities
        a.set_tags([("NM", nm, "i")])

        other_alignments.append(a)

    return other_alignments

def read_bam(contig, start, end, options):
    """Parse BAM record to extract SVs."""
    # bam = pysam.AlignmentFile(options.bam, threads=options.num_threads)
    from svpg.main import bam
    sv_signatures, sv_signatures_inter = [], []
    for current_alignment in bam.fetch(contig, start, end):
        try:
            if current_alignment.is_unmapped or current_alignment.is_secondary or current_alignment.mapping_quality < options.min_mapq or current_alignment.reference_start < start:
                continue
            sigs = decompose_cigars(current_alignment, bam, current_alignment.query_name, options.min_sv_size)
            if sigs:
                sigs = merge_cigar(sigs, options.read)
                sv_signatures.extend(sigs)
            if not current_alignment.is_supplementary:
                supplementary_alignments = retrieve_other_alignments(current_alignment, bam)
                good_suppl_alns = [aln for aln in supplementary_alignments if
                                   not aln.is_unmapped and aln.mapping_quality >= options.min_mapq]
                sig_list = decompose_split(current_alignment, good_suppl_alns, bam)
                sv_signatures_inter.extend(sig_list)

        except StopIteration:
            break
        except KeyboardInterrupt:
            logging.warning('Execution interrupted by user. Stop detection and continue with next step..')
            break

    sv_signatures = sv_signatures + sv_signatures_inter

    return sv_signatures



from random import sample
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster

def form_bins(sv_signatures, max_distance):
    """Form partitions of signatures using mean distance."""
    sorted_signatures = sorted(sv_signatures, key=lambda evi: evi.get_key())
    grouped_bin, current_group = [], []
    for i, sv_sig in enumerate(sorted_signatures):
        if i == 0:
            current_group.append(sv_sig)
        else:
            if current_group[-1].downstream_distance_to(sv_sig) < max_distance:
                current_group.append(sv_sig)
            else:
                grouped_bin.append(current_group)
                current_group = [sv_sig]
    if current_group:
        grouped_bin.append(current_group)

    bin_depth = [len(bin) for bin in grouped_bin]
    mean_depth = sum(bin_depth) / len(bin_depth)

    return grouped_bin, mean_depth

def span_position_distance(signature1, signature2, type, param):
    if type == 'BND':
        if signature1.get_source()[0] != signature2.get_source()[0] or signature1.get_destination()[0] != signature2.get_destination()[0]:
            return 999
        dist1 = abs(signature1.get_source()[1] - signature2.get_source()[1])
        dist2 = abs(signature1.get_destination()[1] - signature2.get_destination()[1])
        if max(dist1, dist2) == 0:
            return 0
        return dist1 // (max(dist1, dist2) * param) + dist2 // (max(dist1, dist2) * param)
    else:
        span1 = signature1.get_source()[2] - signature1.get_source()[1]
        span2 = signature2.get_source()[2] - signature2.get_source()[1]
        center1 = (signature1.get_source()[1] + signature1.get_source()[2]) // 2
        center2 = (signature2.get_source()[1] + signature2.get_source()[2]) // 2
        if signature1.node_ls and signature2.node_ls:
            path1, path2 = signature1.node_ls, signature2.node_ls
            node_distance = 0.05 * (1-len(set(path1).intersection(set(path2))) / len(set(path1).union(set(path2))))
            # node_distance = len(set(path1).symmetric_difference(set(path2))) / max(len(path1), len(path2)) * 0.1
        else:
            node_distance = 0
        span_distance = abs(span1 - span2) / max(span1, span2)
        position_distance = abs(center1 - center2)//(max(span1, span2) * param)

    return position_distance + span_distance + node_distance

def cluster_data(bins, mean_len):
    clusters_final = []
    for bin in bins:
        if len(bin) == 1:
            clusters_final.append(bin)
            continue
        elif len(bin) > 100:
            partition_sample = sample(bin, 100)
        else:
            partition_sample = bin
        element_type = partition_sample[0].type
        param = (abs(len(partition_sample) - mean_len)) / max(len(partition_sample), mean_len)+mean_len
        distance_data = []
        for i in range(len(partition_sample) - 1):
            for j in range(i + 1, len(partition_sample)):
                distance_data.append(
                    span_position_distance(partition_sample[i], partition_sample[j], element_type, param))
        Z = linkage(np.array(distance_data), method="average")
        cluster_indices = list(fcluster(Z, 0.3, criterion='distance'))
        new_clusters = [[] for i in range(max(cluster_indices))]
        for signature_index, cluster_index in enumerate(cluster_indices):
            new_clusters[cluster_index - 1].append(partition_sample[signature_index])
        clusters_final.extend(new_clusters)

    return clusters_final



from math import log10
import numpy as np
# import pysam

err = 0.2
prior = float(1/3)
Genotype = ["0/0", "0/1", "1/1"]

def threshold_ref_count(num):
    if num <= 2:
        return 20*num
    elif 3 <= num <= 5:
        return 9*num
    elif 6 <= num <= 15:
        return 7*num
    else:
        return 5*num

def log10sumexp(log10_probs):
    # Normalization of Genotype likelihoods
    m = max(log10_probs)
    return m + log10(sum(pow(10.0, x-m) for x in log10_probs))

def normalize_log10_probs(log10_probs):
    # Adjust the Genotype likelihoods
    log10_probs = np.array(log10_probs)
    lse = log10sumexp(log10_probs)
    return np.minimum(log10_probs - lse, 0.0)

def rescale_read_counts(c0, c1, max_allowed_reads=100):
    """Ensures that n_total <= max_allowed_reads, rescaling if necessary."""
    Total = c0 + c1
    if Total > max_allowed_reads:
        c0 = int(max_allowed_reads * float(c0/Total))
        c1 = max_allowed_reads - c0
    return c0, c1

def cal_GL(c0, c1, type, platform):
    if platform == "hifi":
        err = 0.05 if type == "INS" else 0.05
    else:
        err = 0.1 if type == "INS" else 0.1
    # Approximate adjustment of events with larger read depth
    c0, c1 = rescale_read_counts(c0, c1)

    ori_GL00 = np.float64(pow((1 - err), c0) * pow(err, c1) * (1 - prior) / 2)
    ori_GL11 = np.float64(pow(err, c0) * pow((1 - err), c1) * (1 - prior) / 2)
    ori_GL01 = np.float64(pow(0.5, c0 + c1) * prior)

    # normalized genotype likelihood
    prob = list(normalize_log10_probs([log10(ori_GL00), log10(ori_GL01), log10(ori_GL11)]))
    GL_P = [pow(10, i) for i in prob]
    PL = [int(np.around(-10 * log10(i))) for i in GL_P]
    GQ = [int(-10 * log10(GL_P[1] + GL_P[2])), int(-10 * log10(GL_P[0] + GL_P[2])), int(-10 * log10(GL_P[0] + GL_P[1]))]
    QUAL = abs(np.around(-10 * log10(GL_P[0]), 1))

    return Genotype[prob.index(max(prob))], "%d,%d,%d" % (PL[0], PL[1], PL[2]), max(GQ), QUAL

def genotype(candidates, type, options):
    # bam = pysam.AlignmentFile(options.bam, threads=options.num_threads)
    from svpg.main import bam
    for nr, candidate in enumerate(candidates):
        reads_supporting_variant = set(candidate.members)
        max_bias = 1000
        contig, start, end = candidate.get_source()
        contig_length = bam.get_reference_length(contig)
        try:
            alignment_it = bam.fetch(contig=contig, start=max(0, start-max_bias), stop=min(contig_length, end+max_bias))
        except ValueError:
            continue
        #  Count reads that overlap the locus and therefore support the reference
        aln_no = 0
        reads_supporting_reference = set()

        while aln_no < 500:
            try:
                current_alignment = next(alignment_it)
            except StopIteration:
                break
            if current_alignment.query_name in reads_supporting_variant:
                continue
            if current_alignment.is_unmapped or current_alignment.is_secondary or current_alignment.mapping_quality < options.min_mapq:
                continue
            aln_no += 1
            if type == "DEL":
                minimum_overlap = min((end - start) / 2, 2000)
                if (current_alignment.reference_start < (end - minimum_overlap) and current_alignment.reference_end > (end + max_bias) or
                    current_alignment.reference_start < (start - max_bias) and current_alignment.reference_end > (start + minimum_overlap)):
                    reads_supporting_reference.add(current_alignment.query_name)
            else:
                if current_alignment.reference_start < (start - max_bias) and current_alignment.reference_end > (end + max_bias):
                    reads_supporting_reference.add(current_alignment.query_name)

        GT, GL, GQ, QUAL = cal_GL(len(reads_supporting_reference), len(reads_supporting_variant), type, options.read)

        candidate.support_fraction = len(reads_supporting_variant) / (
                len(reads_supporting_variant) + len(reads_supporting_reference))
        candidate.genotype = GT
        candidate.ref_reads = len(reads_supporting_reference)
        candidate.alt_reads = len(reads_supporting_variant)

    return candidates


import os
import subprocess
import time

import pysam.bcftools

def process_fasta(input_file, output_file):
    with open(input_file, 'r') as infile, open(output_file, 'w') as outfile:
        total_written = 0
        currently_written = 0
        previous_label = ""
        nr_blocks = 0
        prev_n = False
        prev_break = True

        for line in infile:
            if line == "":
                continue
            if line.startswith(">"):
                previous_label = line.strip().split()[0]
                nr_blocks = 0
                prev_n = True
                continue
            for c in line:
                if c == '\n':
                    continue
                if c in ['n', 'N']:
                    prev_n = True
                    continue
                elif prev_n:
                    if not prev_break:
                        outfile.write("\n")
                    outfile.write(previous_label + "_" + str(nr_blocks) + "\n" + c)
                    total_written += 1
                    prev_break = False
                    currently_written = 1
                    nr_blocks += 1
                    prev_n = False

                else:
                    if currently_written % 70 == 0:
                        outfile.write("\n")
                        prev_break = True
                    outfile.write(c)
                    prev_break = False
                    currently_written += 1
                    total_written += 1
        if not prev_break:
            outfile.write("\n")
    print("Wrote " + str(total_written) + " non-N bases to output file.")


def augment_pipe(base_dir, ref_file, pan_file, output_file):
    os.chdir(base_dir)
    os.makedirs('tmp', exist_ok=True)

    print(">>> Merge VCF files using bcftools.")
    # subprocess.run("bcftools merge -l filelist.tsv --missing-to-ref --force-samples  -W -Oz -o merged.vcf.gz", shell=True, check=True)
    pysam.bcftools.merge("-l", "filelist.tsv", "--missing-to-ref", "--force-samples", "-Oz", "-o", "tmp/merged.vcf.gz", force=True, catch_stdout=False)
    pysam.tabix_index("tmp/merged.vcf.gz", preset="vcf", force=True)
    # subprocess.run("tabix -f merged.vcf.gz", shell=True, check=True)

    print(">>> Normalize the merged VCF file to ensure biallelic representation.")
    subprocess.run("bcftools norm -m- tmp/merged.vcf.gz  -Oz -o tmp/merged_biallelic.vcf.gz", shell=True, check=True)
    # pysam.bcftools.norm("-m-", "tmp/merged.vcf.gz", "-Oz", "-o", "tmp/merged_biallelic.vcf.gz", catch_stdout=False)
    pysam.tabix_index("tmp/merged_biallelic.vcf.gz", preset="vcf", force=True)

    print(">>> Collapse the normalized VCF file using Truvari.")
    collapse_cmd = f"truvari collapse -i tmp/merged_biallelic.vcf.gz -o tmp/collapsed.tmp -c tmp/output.collapsed -f {ref_file}"
    subprocess.run(collapse_cmd, shell=True, check=True)
    # pysam.bcftools.sort("tmp/collapsed.tmp", "-Oz", "-o", "variants.vcf.gz", force=True, catch_stdout=False)
    subprocess.run("bcftools sort tmp/collapsed.tmp -W -Oz -o variants.vcf.gz", shell=True, check=True)
    # time.sleep(1)
    # if os.path.exists("variants.vcf.gz.csi"):
    #     os.remove("variants.vcf.gz.csi")
    # pysam.tabix_index("variants.vcf.gz", preset="vcf", force=True)

    print(">>> Generate the consensus sequence from the VCF file.")
    # pysam.bcftools.consensus("-f", ref_file, "variants.vcf.gz", "-o", "tmp/cons.fa", force=True, catch_stdout=False)
    # subprocess.run("sed 's/^>/&augment_/' tmp/cons.fa > tmp/cons.tmp", shell=True, check=True)
    subprocess.run(f"bcftools consensus -f {ref_file} variants.vcf.gz | sed 's/^>/&augment_/' > tmp/cons.tmp", shell=True, check=True)

    process_fasta('tmp/cons.tmp', 'cons_noN.fa')
    subprocess.run("rm -rf tmp/*", shell=True, check=True)

    print(f">>> Construct a augment graph using Minigraph to {output_file}")
    subprocess.run(f"minigraph -cxggs -t128 {pan_file} cons_noN.fa > {output_file}", shell=True, check=True)


import re
from collections import defaultdict

from svpg.SVSignature import SignatureDeletion, SignatureInsertion
from svpg.util import analyze_cigar_indel, merge_cigar

CIGAR_PATTERN = re.compile(r'(\d+)([MIDNSHP=X])')

class Gaf:
    def __init__(self):
        self.query_name = ""
        self.query_length = 0
        self.query_start = 0
        self.query_end = 0
        self.strand = ""
        self.path = ""
        self.path_length = 0
        self.path_start = 0
        self.path_end = 0
        self.mapping_quality = 0
        self.is_primary = True
        self.cigar = ""
        self.ds = ""

def parse_gaf_line(tokens, gfa_node):
    """Parse a single GAF line"""
    gafline = Gaf()

    if '@' in tokens[0]:
        bam_tags = tokens[0].split('@')
        gafline.query_name = bam_tags[0]
        gafline.type = bam_tags[1]
        gafline.pos = bam_tags[2]
    else:
        gafline.query_name = tokens[0]

    gafline.query_length = int(tokens[1])
    gafline.query_start = int(tokens[2])
    gafline.query_end = int(tokens[3])
    gafline.path = re.findall(r'[<>][^<>]+', tokens[5])
    gafline.strand = '+' if gafline.path[0][0] == '>' else '-'
    # gafline.strand = tokens[4]
    gafline.contig = gfa_node[gafline.path[0][1:]].contig
    gafline.offset = gfa_node[gafline.path[0][1:]].offset
    gafline.path_length = int(tokens[6])
    gafline.path_start = int(tokens[7])
    gafline.path_end = int(tokens[8])
    gafline.mapping_quality = int(tokens[11])
    gafline.is_primary = True

    for tok in tokens:
        if "tp:A:" in tok:
            if tok[5:7] != "P":
                gafline.is_primary = False
        if "cg:Z:" in tok:
            gafline.cigar = tok[5:]
        if "ds:Z:" in tok:
            gafline.ds = tok[6:]

    return gafline

def decompose_split(g_list):
    """Parse GAF record to extract SVs from split_reads."""
    alignment_list, split_signature = [], []
    for g in g_list:
        strand = g.strand
        ref_start = g.offset + g.path_start
        ref_end = g.offset + g.path_end
        q_start = g.query_start
        q_end = g.query_end
        alignment_dict = {
            'read_name': g.query_name,
            'q_start': q_start,
            'q_end': q_end,
            'ref_id': g.contig,
            'ref_start': ref_start,
            'ref_end': ref_end,
            'is_reverse': strand == '-'
        }
        alignment_list.append(alignment_dict)
    sorted_alignment_list = sorted(alignment_list, key=lambda aln: (aln['q_start'], aln['q_end']))
    ultra_ins_flag = True if len(alignment_list) >= 3 and sorted_alignment_list[0]['ref_id'] != sorted_alignment_list[1]['ref_id'] else False

    for alignment_current, alignment_next in zip(sorted_alignment_list[:-1], sorted_alignment_list[1:]):
        distance_on_read = alignment_next['q_start'] - alignment_current['q_end']
        if not alignment_current['is_reverse']:
            distance_on_reference = alignment_next['ref_start'] - alignment_current['ref_end']
            if alignment_next['is_reverse']:  # INV:+-
                if alignment_current['ref_end'] > alignment_next['ref_end']:
                    distance_on_reference = alignment_next['ref_end'] - alignment_current['ref_start']
                else:
                    distance_on_reference = alignment_current['ref_end'] - alignment_next['ref_start']
        else:
            continue
            # distance_on_reference = alignment_current['ref_start'] - alignment_next['ref_end']
            # if not alignment_next['is_reverse']:  # INV:-+
            #     if alignment_current['ref_end'] > alignment_next['ref_end']:
            #         distance_on_reference = alignment_next['ref_end'] - alignment_current['ref_start']
            #     else:
            #         distance_on_reference = alignment_current['ref_end'] - alignment_next['ref_start']
        deviation = distance_on_read - distance_on_reference

        if alignment_current['ref_id'] == alignment_next['ref_id']:
            if alignment_current['is_reverse'] == alignment_next['is_reverse']:
                if distance_on_reference >= -50:
                    if deviation >= 50:  # INS
                        if not alignment_current['is_reverse']:
                            if not ultra_ins_flag:
                                start = alignment_current['ref_end']
                            else:
                                start = min(alignment_current['ref_end'], alignment_next['ref_start'])
                        else:
                            if not ultra_ins_flag:
                                start = alignment_current['ref_start']
                            else:
                                start = min(alignment_current['ref_start'], alignment_next['ref_end'])
                        split_signature.append(
                            SignatureInsertion(alignment_current['ref_id'], start, deviation, "suppl",
                                               alignment_current['read_name'], alt_seq='<INS>',
                                               qry_start=alignment_current['q_end'], qry_end=alignment_next['q_start']))
                    elif deviation <= -50:  # DEL
                        if not alignment_current['is_reverse']:
                            start = alignment_current['ref_end']
                        else:
                            start = alignment_next['ref_end']
                        split_signature.append(
                            SignatureDeletion(alignment_current['ref_id'], start, -deviation, "suppl", alignment_current['read_name']))
                    else:
                        continue

    return split_signature

def pan_node_offset(pan_node, node_list, gfa_node):
    """ Find contig and coordinates for pan_node according to linear_node """
    pan_len = 0
    for node in node_list[node_list.index(pan_node):]:
        node_id = node[1:]
        if gfa_node[node[1:]].sr == 0:
            node_contig = gfa_node[node_id].contig
            node_offset = gfa_node[node_id].offset - pan_len
            return (node_contig, node_offset)
        pan_len += gfa_node[node_id].len
    else:
        return None

def check_continuity(node_list):
    node_num = [int(s[2:]) for s in node_list][::-1]
    if all(node_num[i] - node_num[i - 1] == 1 for i in range(1, len(node_num))):
        return False
    else:
        return node_list

def extract_tsd_alt(ds_seq):
    # ds_seq: '[aatttttgtattt]ttaa...'
    pattern = r'(?:\[([^\[\]]+)\])?([^\[\]]*)(?:\[([^\[\]]+)\])?'
    match = re.fullmatch(pattern, ds_seq)
    if match:
        tsdl, alt, tsdr = match.groups()
        return tsdl or "", alt or "", tsdr or ""
    else:
        return "", ds_seq, ""

def get_node_index_for_pos(pos, cum_lengths):
    for i in range(len(cum_lengths)-1):
        if cum_lengths[i] <= pos < cum_lengths[i+1]:
            return i
    return None

def decompose_cigars(g, gfa_node, min_indel_length=40):
    sigs = []
    node_list = g.path  # ['>s1','>s2']
    first_node = node_list[0]
    first_node_len = gfa_node[first_node[1:]].len
    if gfa_node[first_node[1:]].sr != 0:
        node_result = pan_node_offset(first_node, node_list, gfa_node)
        if node_result:
            g.contig, g.offset = node_result
        else:  # node list is composed entirely of pan_nodes
            return []

    parsed_cigar = CIGAR_PATTERN.findall(g.cigar)
    cigar_tuple = [(int(length), operation) for length, operation in parsed_cigar]
    vars = analyze_cigar_indel(cigar_tuple, min_indel_length, is_gaf=True)
    if vars:
        DS_PATTERN = re.compile(rf'[+-]([atcgn\[\]]{{{min_indel_length},}})', re.IGNORECASE)
        parsed_ds = DS_PATTERN.findall(g.ds) if g.ds else []
    else:
        return []

    effective_first_len = first_node_len - g.path_start
    cum_lengths = [0, effective_first_len]
    for node in node_list[1:]:
        node_len = gfa_node[node[1:]].len
        cum_lengths.append(cum_lengths[-1] + node_len)

    ref_chr = g.contig
    global_ref = g.offset + g.path_start
    last_found_node_index = 0
    for var_index, (pos_ref, pos_read, length, typ) in enumerate(vars):
        ds_str = parsed_ds[var_index] if parsed_ds else ''
        ltsd, alt_seq, rtsd = extract_tsd_alt(ds_str)
        ltsd_len, alt_len, rtsd_len = len(ltsd), len(alt_seq), len(rtsd)

        if (ltsd or rtsd) and length < 1000:  # TSD
            indel_left = pos_ref - ltsd_len
            indel_right = pos_ref + alt_len + rtsd_len

            left_node_index = get_node_index_for_pos(indel_left, cum_lengths)
            right_node_index = get_node_index_for_pos(indel_right - 1, cum_lengths)

            if left_node_index != right_node_index:  # Filter cross-node indels
                continue

        start = None
        if g.strand == '+':  # first node is forward
            start = global_ref + pos_ref
        else:
            # if check_continuity(node_list):
            #     return []
            if pos_ref < first_node_len - g.path_start:  # the indel is in first node
                global_ref = g.offset + (first_node_len - g.path_start)
                if typ == 'INS':
                    start = global_ref - pos_ref
                else:
                    start = global_ref - pos_ref - length
            else:  # find global_ref according to node offset
                start_index = max(0, last_found_node_index)
                for i in range(start_index, len(node_list)):
                    node = node_list[i]
                    node_name = node[1:]
                    node_len = gfa_node[node_name].len

                    node_start_ref = cum_lengths[i]
                    node_end_ref = cum_lengths[i + 1]

                    if node_start_ref <= pos_ref < node_end_ref:
                        last_found_node_index = i

                        if gfa_node[node_name].sr == 0:  # the node is a linear node
                            global_ref = gfa_node[node_name].offset
                        else:  # the node is a pan node
                            node_result = pan_node_offset(node, node_list, gfa_node)
                            if not node_result:
                                break
                            else:
                                global_ref = node_result[1]

                        local_ref = pos_ref - node_start_ref
                        if node[0] == '>':  # the node is forward
                            start = global_ref + local_ref
                        else:  # the node is reverse
                            if typ == "INS":
                                start = global_ref + node_len - local_ref
                            else:
                                start = global_ref + node_len - local_ref - length
                        break
        if start is None:
            continue

        if typ == "DEL":
            sigs.append(SignatureDeletion(ref_chr, start, length, "cigar", g.query_name))
        elif typ == "INS":
            sigs.append(SignatureInsertion(ref_chr, start, length, "cigar", g.query_name, alt_seq=ltsd+alt_seq+rtsd))

    return sigs


def read_gaf(gaf_chunk, gfa_node, options):
    """Parse SVsignatures GAF record to extract SVs."""
    inconsist_read = []
    sv_signatures = []
    read_dict = {}
    j, k, e = 0, 0, 0
    min_sv_size = options.min_sv_size
    for i, line in enumerate(gaf_chunk):
        tokens = line.strip().split('\t')
        if tokens[4] == '*':
            continue

        g = parse_gaf_line(tokens, gfa_node)
        if g.mapping_quality < options.min_mapq:
            j += 1
            continue

        if tokens[0] in read_dict:
            read_dict[tokens[0]].append(g)
        else:
            read_dict[tokens[0]] = [g]

        node_list = g.path  # ['>s1','>s2','>s3']
        node_sr = [gfa_node[node[1:]].sr for node in node_list]
        sigs = []

        if sum(node_sr) == 0:
            for node_current, node_next in zip(node_list[:-1], node_list[1:]):
                # map to non-adjacent nodes, ['>s1', '>s3']
                split_node_temp = list(range(min(int(node_current[2:]), int(node_next[2:])) + 1,
                                             max(int(node_current[2:]), int(node_next[2:]))))
                if len(split_node_temp) > 0:
                    if node_current[0] == '>':  # ['>s1', '>s3']
                        start = gfa_node[node_current[1:]].offset + gfa_node[node_current[1:]].len
                        end = gfa_node[node_next[1:]].offset
                    else:  # ['<s3', '>s1']
                        start = gfa_node[node_next[1:]].offset
                        end = gfa_node[node_current[1:]].offset
                    sigs.append(SignatureDeletion(g.contig, start, end - start, "ref_split", g.query_name, node_ls=split_node_temp))

            sigs_cigar = decompose_cigars(g, gfa_node, min_sv_size)
            sigs.extend(sigs_cigar)
        else:
            sigs_liner_pan = decompose_cigars(g, gfa_node, min_indel_length=10)
            sigs_cigar = [sig for sig in sigs_liner_pan if sig.svlen >= min_sv_size]

            linear_index = [i for i, x in enumerate(node_sr) if x == 0]
            linear_node = [node_list[i] for i in linear_index]
            for node_current, node_next in zip(linear_node[:-1], linear_node[1:]):
                if node_current[0] == '>':
                    start = gfa_node[node_current[1:]].offset + gfa_node[node_current[1:]].len
                    # Only retaining the cigar SVs of the linear nodes
                    sigs_cigar = [sig for sig in sigs_cigar if
                                  not (start <= sig.start <= gfa_node[node_next[1:]].offset)]
                else:
                    start = gfa_node[node_next[1:]].offset
                    sigs_cigar = [sig for sig in sigs_cigar if
                                  not (start <= sig.start <= gfa_node[node_current[1:]].offset)]
                split_node_temp = list(range(min(int(node_current[2:]), int(node_next[2:])) + 1,
                                             max(int(node_current[2:]), int(node_next[2:]))))
                split_len = sum([gfa_node['s' + str(node)].len for node in split_node_temp])
                pan_node = node_list[node_list.index(node_current) + 1:node_list.index(node_next)]

                # map to a pan node: insersion
                if pan_node:
                    length = sum([gfa_node[pan[1:]].len for pan in pan_node])
                    for indel in sigs_liner_pan:
                        if start <= indel.start <= start + length:  # cigar SVs in the pan node
                            if indel.type == "DEL":
                                length -= indel.svlen
                            else:
                                length += indel.svlen
                    if length - split_len >= min_sv_size:
                        alt_seq = ''.join([gfa_node[node[1:]].sequence for node in pan_node])
                        sigs.append(SignatureInsertion(g.contig, start, length - split_len, "ref_split", g.query_name, alt_seq=alt_seq, node_ls=pan_node))

                if len(split_node_temp) > 0:  # map to a missing linear nodes: deletion
                    if sum([gfa_node['s' + str(node)].len for node in split_node_temp]) < min_sv_size:
                        continue
                    if node_current[0] == '>':
                        end = gfa_node[node_next[1:]].offset
                    else:
                        end = gfa_node[node_current[1:]].offset
                    if pan_node and length >= min_sv_size:  # the pan node inserted in the middle
                        end -= length
                    if end - start >= min_sv_size:
                        sigs.append(SignatureDeletion(g.contig, start, end - start, "ref_split", g.query_name))

            sigs = sigs+sigs_cigar

        sigs_initial = [sig for sig in sigs if sig.type == g.type]
        sigs = merge_cigar(sigs_initial, options.read)

        bam_pos = g.pos.split(':')
        bam_len = int(bam_pos[2]) - int(bam_pos[1])
        sigs_ = []
        # Find the closest SV record
        if len(sigs) > 1:
            min_distance = float("inf")
            for sig in sigs:
                sig_distance = abs(int(bam_pos[1]) - sig.start) + abs(int(bam_pos[2]) - sig.end)
                if sig_distance < min_distance:
                    min_distance = sig_distance
                    sigs_ = [sig]
        elif len(sigs) == 1:
            sigs_ = sigs

        if sigs_ and min(sigs_[0].svlen, bam_len) / max(sigs_[0].svlen, bam_len) < 0.5:  # todo:remap
            if tokens[0].split('@')[-1] == 'merged_indel':
                pass
            else:
                inconsist_read.append(tokens[0])
                k += 1
                continue

        sv_signatures.extend(sigs)

    for key, value in read_dict.items():
        if len(value) > 1:
            e += 1
            var_split = decompose_split(value)
            sv_signatures.extend(var_split)

    print(f"low_mapq: {j}, inconsistent_read: {k},  split_read: {e}")

    return sv_signatures


def read_gaf_pan(gaf_chunk, gfa_node, options):
    """Parse WGS GAF record to extract SVs."""
    sv_signatures = []
    read_dict = defaultdict(list)

    for line in gaf_chunk:
        tokens = line.strip().split('\t')
        if tokens[4] == '*':
            continue
        g = parse_gaf_line(tokens, gfa_node)
        if g.mapping_quality < options.min_mapq:
            continue

        read_dict[tokens[0]].append(g)

        if g.query_end - g.query_start < g.query_length * 0.7:  # filter cigar in short alignments
            continue

        sigs = decompose_cigars(g, gfa_node, options.min_sv_size)
        if len(sigs) > 1 and len(sigs) > g.query_length * 1e-4 * 2:
            continue
        sigs_merged = merge_cigar(sigs, options.read)
        sv_signatures.extend(sigs_merged)

    for key, value in read_dict.items():
        if len(value) > 1:
            var_split = decompose_split(value)
            sv_signatures.extend(var_split)

    return sv_signatures




import sys
import os
import argparse


def parse_arguments(arguments=sys.argv[1:]):
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="""SVPG - Structural variant detection based on pangenome graph""")
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    subparsers = parser.add_subparsers(help='mode', dest='sub')
    parser.set_defaults(sub='call')

    parser_bam = subparsers.add_parser('call',
                                       help='Pangenome-guided SV detection')
    parser_bam.add_argument('--working_dir',
                            type=os.path.abspath,
                            help='Working and output directory. \
                                  Existing files in the directory are overwritten. \
                                  If the directory does not exist, it is created.')
    parser_bam.add_argument('--bam',
                            type=str,
                            help='Coordinate-sorted and indexed BAM file with aligned long reads')
    parser_bam.add_argument('--ref',
                            type=str,
                            help='Liner reference genome file that the long reads were aligned to (.fa)')
    parser_bam.add_argument('-o', '--out',
                            type=str,
                            default='variants.vcf',
                            help='VCF output file name')
    parser_bam.add_argument('--gfa',
                            type=str,
                            help='Pangenome reference file that the long reads were aligned to (.gfa)')
    parser_bam.add_argument('-t', '--num_threads',
                            type=int,
                            default=16,
                            help='Number of threads to use')
    parser_bam.add_argument('--read',
                            type=str,
                            default='ont',
                            help='Platform type for sequencing data, either "ont", "hifi" or "clr" (default: %(default)s)')
    parser_bam.add_argument('--min_mapq',
                            type=int,
                            default=20,
                            help='Minimum mapping quality of reads to consider (default: %(default)s). \
                                  Reads with a lower mapping quality are ignored.')
    parser_bam.add_argument('--min_sv_size',
                            type=int,
                            default=40,
                            help='Minimum SV size to detect (default: %(default)s). \
                                  SVPG can potentially detect events of any size but is limited by the \
                                  signal-to-noise ratio in the input alignments. That means that more \
                                  accurate reads and alignments enable the detection of smaller events. \
                                  For current PacBio or Nanopore data, we would recommend a minimum size \
                                  of min_sv_sizebp or larger.')
    parser_bam.add_argument('--max_sv_size',
                            type=int,
                            default=-1,
                            help='Maximum SV size to detect include sequence information. Set to -1 for unlimited size.')
    parser_bam.add_argument('--noseq',
                            action='store_true',
                            default=False,
                            help='Omit sequence information in VCF. Recommended for somatic SV analysis with many large SVs.')
    parser_bam.add_argument('-d', '--depth',
                            type=int,
                            help='Sequencing depth of this dataset')
    parser_bam.add_argument('-s', '--min_support',
                            type=int,
                            help='Minimal number of supporting reads for one SV event')
    parser_bam.add_argument('--types',
                            type=str,
                            default="DEL,INS",
                            help='SV types to include in output VCF (default: %(default)s). \
                                  Give a comma-separated list of SV types. The possible SV types are: DEL (deletions), \
                                  INS (novel insertions)')
    parser_bam.add_argument('--contigs',
                            type=str,
                            nargs='*',
                            help='Specify the chromosomes list to call SVs (e.g., --contigs chr1 chr2 chrX)')
    parser_bam.add_argument('--skip_genotype',
                            action='store_true',
                            help='Skip genotyping and only call SVs')

    ##########################################################
    parser_gaf = subparsers.add_parser('graph-call',
                                       help='Pangenome-based de novo SV detection')
    parser_gaf.add_argument('--working_dir',
                            type=os.path.abspath,
                            help='Working and output directory. \
                                  Existing files in the directory are overwritten. \
                                  If the directory does not exist, it is created.')
    parser_gaf.add_argument('--ref',
                            type=str,
                            help='Liner reference genome file that the long reads were aligned to (.fa)')
    parser_gaf.add_argument('--gfa',
                            type=str,
                            help='Pangenome reference file that the long reads were aligned to (.gfa)')
    parser_gaf.add_argument('--gaf',
                            type=str,
                            help='GAF file that aligns to the pangenome reference (.gaf)')
    parser_gaf.add_argument('-o', '--out',
                            type=str,
                            default='variants.vcf',
                            help='VCF output file name')
    parser_gaf.add_argument('-t', '--num_threads',
                            type=int,
                            default=16,
                            help='Number of threads to use')
    parser_gaf.add_argument('--read',
                            type=str,
                            default='ont',
                            help='Platform type for sequencing data, either "ont", "hifi" or "clr"  (default: %(default)s)')
    parser_gaf.add_argument('--raw_fasta',
                            type=str,
                            help='Raw fasta file that the long reads. Since the GAF format does not store \
                                  complete sequences information, providing the original reads\
                                  sequencing data through the --raw_fasta parameter becomes essential \
                                  in graph augmentation mode or when forced output of ALT information \
                                  for split alignments is required. ')
    parser_gaf.add_argument('--min_mapq',
                            type=int,
                            default=20,
                            help='Minimum mapping quality of reads to consider (default: %(default)s). \
                                  Reads with a lower mapping quality are ignored.')
    parser_gaf.add_argument('--min_sv_size',
                            type=int,
                            default=40,
                            help='Minimum SV size to detect (default: %(default)s). \
                                  SVPG can potentially detect events of any size but is limited by the \
                                  signal-to-noise ratio in the input alignments. That means that more \
                                  accurate reads and alignments enable the detection of smaller events. \
                                  For current PacBio or Nanopore data, we would recommend a minimum size \
                                  of 40bp or larger.')
    parser_gaf.add_argument('--max_sv_size',
                            type=int,
                            default=-1,
                            help='Maximum SV size to detect include sequence information. Set to -1 for unlimited size.')
    parser_gaf.add_argument('--noseq',
                            action='store_true',
                            default=False,
                            help='Omit sequence information in VCF. Recommended for somatic SV analysis with many large SVs.')
    parser_gaf.add_argument('-d', '--depth',
                            type=int,
                            help='Sequencing depth of this dataset')
    parser_gaf.add_argument('-s', '--min_support',
                            type=int,
                            help='Minimal number of supporting reads for one SV event')
    parser_gaf.add_argument('--types',
                            type=str,
                            default="DEL,INS",
                            help='SV types to include in output VCF (default: %(default)s). \
                                  Give a comma-separated list of SV types. The possible SV types are: DEL (deletions), \
                                  INS (novel insertions)')
    parser_gaf.add_argument('--contigs',
                            type=str,
                            nargs='*',
                            help='Specify the chromosomes list to call SVs (e.g., --contigs chr1 chr2 chrX)')

    ##########################################################
    parser_augment = subparsers.add_parser('augment',
                                           help='Pangenome graph augmentation pipeline')
    parser_augment.add_argument('--working_dir',
                                type=os.path.abspath,
                                help='Working and output directory. \
                                      Existing files in the directory are overwritten. \
                                      If the directory does not exist, it is created.')
    parser_augment.add_argument('--ref',
                                type=str,
                                help='Liner reference genome file that the long reads were aligned to (.fa)')
    parser_augment.add_argument('--gfa',
                                type=str,
                                help='Pangenome reference file that the long reads were aligned to (.gfa)')
    parser_augment.add_argument('-o', '--out',
                                type=str,
                                default='augment.gfa',
                                help='Augmented GFA output file name')
    parser_augment.add_argument('--vcf_out',
                                type=str,
                                default='variants.vcf',
                                help='VCF output file name')
    parser_augment.add_argument('-t', '--num_threads',
                                type=int,
                                default=16,
                                help='Number of threads to use')
    parser_augment.add_argument('--read',
                                type=str,
                                default='ont',
                                help='Platform type for sequencing data, either "ont", "hifi" or "clr"  (default: %(default)s)')
    parser_augment.add_argument('--sample_list',
                                type=str,
                                default='',
                                help='Provide a .tsv file listing the paths to FASTA files of new samples. \
                                For example, the sample.tsv file may look like(sample_1 name ≠ sample_2 name): /path/to/sample_1.fasta\n/path/to/sample_2.fasta')
    parser_augment.add_argument('--min_mapq',
                                type=int,
                                default=20,
                                help='Minimum mapping quality of reads to consider (default: %(default)s). \
                                        Reads with a lower mapping quality are ignored.')
    parser_augment.add_argument('--min_sv_size',
                                type=int,
                                default=40,
                                help='Minimum SV size to detect (default: %(default)s). \
                                      SVPG can potentially detect events of any size but is limited by the \
                                      signal-to-noise ratio in the input alignments. That means that more \
                                      accurate reads and alignments enable the detection of smaller events. \
                                      For current PacBio or Nanopore data, we would recommend a minimum size \
                                      of 40bp or larger.')
    parser_augment.add_argument('--max_sv_size',
                                type=int,
                                default=-1,
                                help='Maximum SV size to detect include sequence information. Set to -1 for unlimited size.')
    parser_augment.add_argument('--skip_call',
                                action='store_true',
                                help='Skip call SVs and only graph augment')

    return parser.parse_args(arguments)



import numpy as np
import time
import os.path
from collections import defaultdict

from svpg.util import sorted_nicely

class Candidate:
    def __init__(self, contig, start, end, sv_from, type, members, ref_seq='N', alt_seq='.', genotype='1/1', ref_reads=None, alt_reads=None):
        self.contig = contig
        self.start = start
        self.end = end
        self.sv_from = sv_from
        self.type = type
        self.members = members
        self.ref_seq = ref_seq
        self.alt_seq = alt_seq
        self.score = len(members)
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads

    def get_source(self):
        return (self.contig, self.start, self.end)

    def get_vcf_entry(self):
        contig, start, end = self.get_source()
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        info_template = "SVTYPE={0};END={1};SVLEN={2};SUPPORT={3}"
        info_string = info_template.format(self.type, end if self.type == "DEL" else start, start - end if self.type == "DEL" else end - start, self.score)
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
            chrom=contig,
            pos=start,
            id="PLACEHOLDERFORID",
            ref=self.ref_seq,
            alt=self.alt_seq,
            qual=self.score,
            filter="PASS" if len(filters) == 0 else ";".join(filters),
            info=info_string,
            format="GT:DP:AD",
            samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".",
                                                   alt=self.alt_reads if self.alt_reads != None else "."))

def consolidate_clusters_unilocal(clusters, ref_chrom, options, fasta_file=None):
    """Consolidate clusters to a list of (type, contig, mean start, mean end, cluster size, members) tuples."""
    min_sv_length, noseqs = options.min_sv_size, options.noseq
    max_sv_length = float('inf') if options.max_sv_size == -1 else options.max_sv_size

    consolidated_clusters = []
    for index, cluster in enumerate(clusters):
        svtype = cluster[0].type
        sv_from = [member.signature for member in cluster]
        members = [member.read_name for member in cluster]
        start = round(np.median([member.get_source()[1] for member in cluster]))
        end = round(np.median([member.svlen for member in cluster])) + start
        svlen = abs(end - start)

        if min_sv_length <= svlen <= max_sv_length:
            if not noseqs:
                try:
                    ref_base = ref_chrom[max(start - 1, 0)]
                    ref_seq = ref_base if svtype == "INS" else ref_chrom[max(start - 1, 0):end]
                except IndexError:
                    continue

                if svtype == "INS":
                    for member in cluster:
                        if member.svlen < svlen:
                            continue
                        if member.alt_seq != "<INS>":
                            alt_seq = ref_base + member.alt_seq
                        else:
                            if fasta_file:
                                alt_seq = ref_base
                            else:
                                alt_seq = "<INS>"
                        break
                else:
                    alt_seq = ref_base if ref_chrom else "<DEL>"
            else:
                ref_seq = "N"
                alt_seq = "<INS>" if svtype == "INS" else "<DEL>"

            consolidated_clusters.append(Candidate(cluster[0].get_source()[0], start, end, sv_from, svtype, members, ref_seq, alt_seq))

    return consolidated_clusters

def write_final_vcf(deletion_candidates,
                    novel_insertion_candidates,
                    contig_names,
                    contig_lengths,
                    types_to_output,
                    working_dir,
                    outfile):
    vcf_output = open(os.path.join(working_dir, outfile), 'w')

    # Write header lines
    print("##fileformat=VCFv4.2", file=vcf_output)
    print("##fileDate={0}".format(time.strftime("%Y-%m-%d|%I:%M:%S%p|%Z|%z")), file=vcf_output)
    for contig_name, contig_length in zip(contig_names, contig_lengths):
        print("##contig=<ID={0},length={1}>".format(contig_name, contig_length), file=vcf_output)
    if "DEL" in types_to_output:
        print("##ALT=<ID=DEL,Description=\"Deletion\">", file=vcf_output)
    if "INS" in types_to_output:
        print("##ALT=<ID=INS,Description=\"Insertion\">", file=vcf_output)
    print("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">", file=vcf_output)
    print("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the variant described in this record\">",
          file=vcf_output)
    print("##INFO=<ID=SVLEN,Number=1,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">",
          file=vcf_output)
    print("##INFO=<ID=SUPPORT,Number=1,Type=Integer,Description=\"Number of reads supporting this variant\">",
          file=vcf_output)
    print("##FILTER=<ID=hom_ref,Description=\"Genotype is homozygous reference\">", file=vcf_output)
    print("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">", file=vcf_output)
    print("##FORMAT=<ID=DP,Number=1,Type=Integer,Description=\"Read depth\">", file=vcf_output)
    print("##FORMAT=<ID=AD,Number=R,Type=Integer,Description=\"Read depth for each allele\">", file=vcf_output)
    print("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tSample", file=vcf_output)

    # Prepare VCF entries depending on command-line parameters
    vcf_entries = []
    if "DEL" in types_to_output:
        for candidate in deletion_candidates:
            vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry(), "DEL"))
    if "INS" in types_to_output:
        for candidate in novel_insertion_candidates:
            vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry(), "INS"))

    # Sort and write entries to VCF
    svtype_counter = defaultdict(int)
    for source, entry, svtype in sorted_nicely(vcf_entries):
        variant_id = "SVPG.{svtype}.{number}".format(svtype=svtype, number=svtype_counter[svtype] + 1)
        entry_with_id = entry.replace("PLACEHOLDERFORID", variant_id, 1)
        svtype_counter[svtype] += 1
        print(entry_with_id, file=vcf_output)

    vcf_output.close()



from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='svpg',
    version='1.1',
    description='A pangenome-based structural variant caller',
    author='henghu',
    author_email='hhengwork@gmail.com',
    license="MIT",
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[
        'numpy>=1.26.4',
        'pysam>=0.22',
        'scipy>=1.13.1'
    ],
    entry_points={
        'console_scripts': [
            'svpg=svpg.main:main',
        ],
    },
    python_requires='>=3.10',
)
